"""Useful constants for experiments and conversions.

Much of this is collected from past students, especially Jeff's 'Cstes.ipf'.

Some of this will disappear in future updates, as we move away from magic constants towards
bundling necessary information on endstation classes.
"""

from numpy import pi
import pint

ureg = pint.UnitRegistry()

# eV, A reasonablish value if you aren't sure for the particular sample
WORK_FUNCTION = 4.3

METERS_PER_SECOND_PER_EV_ANGSTROM = (
    151927  # converts from eV * angstrom to meters/second velocity units
)
HBAR = 1.0545718176461565 * 10 ** (-34)
HBAR_PER_EV = 6.582119569 * 10 ** (
    -16
)  # gives the energy lifetime relationship via tau = -hbar / np.imag(self_energy)
BARE_ELECTRON_MASS = 9.109383701e-31  # kg
HBAR_SQ_EV_PER_ELECTRON_MASS = 0.475600805657  # hbar^2 / m0 in eV^2 s^2 / kg
HBAR_SQ_EV_PER_ELECTRON_MASS_ANGSTROM_SQ = (
    7.619964  # (hbar^2) / (m0 * angstrom ^2) in eV
)

K_BOLTZMANN_EV_KELVIN = 8.617333262145178e-5  # in units of eV / Kelvin
K_BOLTZMANN_MEV_KELVIN = 1000 * K_BOLTZMANN_EV_KELVIN  # meV / Kelvin

HC = 1239.84172  # in units of eV * nm

K_INV_ANGSTROM = 0.5123167219534328
HV_CONVERSION = 3.814697265625

HEX_ALPHABET = "ABCDEF0123456789"
ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
ALPHANUMERIC = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"



"""A lazy keep-alive `multiprocesssing.Pool`.

We keep a pool alive after one is requested at the cost of memory overhead
because otherwise pools are too slow due to heavy analysis imports (scipy, etc.).
"""

from multiprocessing import Pool, pool

# from pathos.pools import ProcessPool
from typing import Optional

__all__ = ["hot_pool"]


class HotPool:
    _pool: Optional[pool.Pool] = None

    @property
    def pool(self) -> pool.Pool:
        if self._pool is not None:
            return self._pool

        self._pool = Pool()
        return self._pool

    def __del__(self):
        if self._pool is not None:
            self._pool.close()
            self._pool = None


hot_pool = HotPool()



"""A widget providing rudimentary information about an axis on a DataArray."""

# pylint: disable=import-error

from PyQt5 import QtWidgets
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .base import ImageTool
    from weakref import ReferenceType

__all__ = ("AxisInfoWidget",)


class AxisInfoWidget(QtWidgets.QGroupBox):
    """A widget providing some rudimentary axis information."""

    def __init__(
        self, parent=None, root: "ReferenceType" = None, axis_name: str = None
    ):
        """Configure inner widgets for axis info, and transpose to front button."""
        super().__init__(title=axis_name, parent=parent)

        self.layout = QtWidgets.QGridLayout(self)

        self.label = QtWidgets.QLabel("Cursor: ")
        self.transpose_button = QtWidgets.QPushButton("To Front")
        self.transpose_button.clicked.connect(self.on_transpose)

        self.layout.addWidget(self.label)
        self.layout.addWidget(self.transpose_button)

        self.axis_name = axis_name
        self._root = root
        self.setLayout(self.layout)
        self.recompute()

    @property
    def root(self) -> "ImageTool":
        """Unwraps the weakref to the parent application."""
        return self._root()

    def recompute(self):
        """Force a recomputation of dependent UI state: here, the title and text."""
        self.setTitle(self.axis_name)
        try:
            cursor_indices = self.root.context["cursor"][self.axis_name]
            self.label.setText(
                f"Cursor: {cursor_indices[0]:.1f}, {cursor_indices[1]:.1f}"
            )
        except KeyError:
            pass

    def on_transpose(self):
        """This UI control lets you tranpose the axis it refers to to the front."""
        try:
            self.root.transpose_to_front(self.axis_name)
        except Exception:
            pass



"""Utilities related to interpretation of model results.

This borrows ideas heavily from fastai which provides interpreter classes
for different kinds of models.
"""

from dataclasses import dataclass, field
import math

import numpy as np
import matplotlib.pyplot as plt

from torch.utils.data import DataLoader
from torch.utils.data.dataset import Subset
import pytorch_lightning as pl
import torch

import tqdm
from typing import List, Any, Optional, Tuple, Union

__all__ = [
    "Interpretation",
    "InterpretationItem",
]


@dataclass
class InterpretationItem:
    """Provides tools to introspect model performance on a single item."""

    target: Any
    predicted_target: Any
    loss: float
    index: int
    parent_dataloader: DataLoader

    @property
    def dataset(self):
        """Fetches the original dataset used to train and containing this item.

        We need to unwrap the dataset in case we are actually dealing
        with a Subset. We should obtain an indexed Dataset at the end
        of the day, and we will know this is the case because we use
        the sentinel attribute `is_indexed` to mark this.

        This may fail sometimes, but this is better than returning junk
        data which is what happens if we get a shuffled view over the
        dataset.
        """
        dset = self.parent_dataloader.dataset

        if isinstance(dset, Subset):
            dset = dset.dataset

        assert dset.is_indexed == True
        return dset

    def show(self, input_formatter, target_formatter, ax=None, pullback=True):
        """Plots this item onto the provided axes. See also the `show` method of `Interpretation`."""
        if ax is None:
            _, ax = plt.subplots()

        dset = self.dataset
        with dset.no_transforms():
            x = dset[self.index][0]

        if input_formatter is not None:
            input_formatter.show(x, ax)

        ax.set_title(
            "Item {index}; loss={loss:.3f}\n".format(
                index=self.index, loss=float(self.loss)
            )
        )

        if target_formatter is not None:
            if hasattr(target_formatter, "context"):
                target_formatter.context = dict(is_ground_truth=True)

            target = self.decodes_target(self.target) if pullback else self.target
            target_formatter.show(target, ax)

            if hasattr(target_formatter, "context"):
                target_formatter.context = dict(is_ground_truth=False)

            predicted = (
                self.decodes_target(self.predicted_target)
                if pullback
                else self.predicted_target
            )
            target_formatter.show(predicted, ax)

    def decodes_target(self, value: Any) -> Any:
        """Pulls the predicted target backwards through the transformation stack.

        Pullback continues until an irreversible transform is met in order
        to be able to plot targets and predictions in a natural space.
        """
        tfm = self.dataset.transforms
        if hasattr(tfm, "decodes_target"):
            return tfm.decodes_target(value)

        return value


@dataclass
class Interpretation:
    """Provides utilities to interpret predictions of a model.

    Importantly, this is not intended to provide any model introspection
    tools.
    """

    model: pl.LightningModule
    train_dataloader: DataLoader
    val_dataloaders: DataLoader

    train: bool = True
    val_index: int = 0

    train_items: List[InterpretationItem] = field(init=False, repr=False)
    val_item_lists: List[List[InterpretationItem]] = field(init=False, repr=False)

    @property
    def items(self) -> List[InterpretationItem]:
        """All of the ``InterpretationItem`` instances inside this instance."""
        if self.train:
            return self.train_items

        return self.val_item_lists[self.val_index]

    def top_losses(self, ascending=False) -> List[InterpretationItem]:
        """Orders the items by loss."""
        key = lambda item: item.loss if ascending else -item.loss
        return sorted(self.items, key=key)

    def show(
        self,
        n_items: Optional[Union[int, Tuple[int, int]]] = 9,
        items: Optional[List[InterpretationItem]] = None,
        input_formatter=None,
        target_formatter=None,
    ) -> None:
        """Plots a subset of the interpreted items.

        For each item, we "plot" its data, its label, and model performance characteristics
        on this item.

        For example, on an image classification task this might mean to plot the image,
        the images class name as a label above it, the predicted class, and the numerical loss.
        """
        layout = None

        if items is None:
            if isinstance(n_items, (tuple, list)):
                layout = n_items
            else:
                n_rows = int(math.ceil(n_items**0.5))
                layout = (n_rows, n_rows)

            items = self.top_losses()[:n_items]
        else:
            n_items = len(items)
            n_rows = int(math.ceil(n_items**0.5))
            layout = (n_rows, n_rows)

        _, axes = plt.subplots(*layout, figsize=(layout[0] * 3, layout[1] * 4))

        items_with_nones = list(items) + [None] * (np.prod(layout) - n_items)
        for item, ax in zip(items_with_nones, axes.ravel()):
            if item is None:
                ax.axis("off")
            else:
                item.show(input_formatter, target_formatter, ax)

        plt.tight_layout()

    @classmethod
    def from_trainer(cls, trainer: pl.Trainer):
        """Builds an interpreter from an instance of a `pytorch_lightning.Trainer`."""
        return cls(trainer.model, trainer.train_dataloader, trainer.val_dataloaders)

    def dataloader_to_item_list(
        self, dataloader: DataLoader
    ) -> List[InterpretationItem]:
        """Converts a data loader into a list of interpretation items corresponding to the data samples."""
        items = []

        for batch in tqdm.tqdm(dataloader.iter_all()):
            x, y, indices = batch
            with torch.no_grad():
                y_hat = self.model(x).cpu()
                y_hats = torch.unbind(y_hat, axis=0)
                ys = torch.unbind(y, axis=0)

                losses = [
                    self.model.criterion(yi_hat, yi) for yi_hat, yi in zip(y_hats, ys)
                ]

            for yi, yi_hat, loss, index in zip(
                ys, y_hats, losses, torch.unbind(indices, axis=0)
            ):
                items.append(
                    InterpretationItem(
                        torch.squeeze(yi),
                        torch.squeeze(yi_hat),
                        torch.squeeze(loss),
                        int(index),
                        dataloader,
                    )
                )

        return items

    def __post_init__(self):
        """Populates train_items and val_item_lists.

        This is done by iterating through the dataloaders and pushing data through the models.
        """
        self.train_items = self.dataloader_to_item_list(self.train_dataloader)
        self.val_item_lists = [
            self.dataloader_to_item_list(dl) for dl in self.val_dataloaders
        ]



"""Utilities related to function application on xr types."""

import numpy as np
import xarray as xr

from typing import Callable, Any
from functools import wraps

from arpes.typing import DataType

__all__ = (
    "apply_dataarray",
    "lift_datavar_attrs",
    "lift_dataarray_attrs",
    "lift_dataarray",
    "lift_spectrum",
    "unwrap_xarray_item",
    "unwrap_xarray_dict",
    "rename_dataset_keys",
)


def unwrap_xarray_item(item):
    """Unwraps something that might or might not be an xarray like with .item() attribute.

    This is especially helpful for dealing with unwrapping coordinates which might
    be floating point-like or might be array-like.

    Args:
        item: The value to unwrap.

    Returns:
        The safely unwrapped item
    """
    try:
        return item.item()
    except (AttributeError, ValueError):
        return item


def unwrap_xarray_dict(d: dict[str, Any]) -> dict[str, Any]:
    """Returns the attributes as unwrapped values rather than item() instances.

    Useful for unwrapping coordinate dicts where the values might be a bare type:
    like a float or an int, but might also be a wrapped array-like for instance
    xr.DataArray. Even worse, we might have wrapped bare values!

    Args:
        d

    Returns:
        The unwrapped attributes as a dict.
    """
    return {k: unwrap_xarray_item(v) for k, v in d.items()}


def apply_dataarray(arr: xr.DataArray, f, *args, **kwargs):
    """Applies a function onto the values of a DataArray."""
    return xr.DataArray(
        f(arr.values, *args, **kwargs), arr.coords, arr.dims, attrs=arr.attrs
    )


def lift_dataarray(
    f: Callable[[np.ndarray], np.ndarray]
) -> Callable[[xr.DataArray], xr.DataArray]:
    """Lifts a function that operates on an np.ndarray's values to act on an xr.DataArray.

    Args:
        f

    Returns:
        g: Function operating on an xr.DataArray
    """

    @wraps(f)
    def g(arr: xr.DataArray, *args, **kwargs):
        return apply_dataarray(arr, f, *args, **kwargs)

    return g


def lift_dataarray_attrs(
    f: Callable[[dict], dict]
) -> Callable[[xr.DataArray], xr.DataArray]:
    """Lifts a function that operates dicts to a function that acts on dataarray attrs.

    Produces a new xr.DataArray.

    Args:
        f

    Returns:
        g: Function operating on the attributes of an xr.DataArray
    """

    @wraps(f)
    def g(arr: xr.DataArray, *args, **kwargs):
        return xr.DataArray(
            arr.values, arr.coords, arr.dims, attrs=f(arr.attrs, *args, **kwargs)
        )

    return g


def lift_datavar_attrs(f: Callable[[dict], dict]) -> Callable[[DataType], DataType]:
    """
    Lifts a function that operates dicts to a function that acts on xr attrs.

    Applies to all attributes of all the datavars in a xr.Dataset, as well as the
    Dataset attrs themselves.

    Args:
        f: Function to apply

    Returns:
        The function modified to apply to xr instances.
    """

    @wraps(f)
    def g(data: DataType, *args, **kwargs):
        arr_lifted = lift_dataarray_attrs(f)
        if isinstance(data, xr.DataArray):
            return arr_lifted(data, *args, **kwargs)

        var_names = list(data.data_vars.keys())
        new_vars = {k: arr_lifted(data[k], *args, **kwargs) for k in var_names}
        new_root_attrs = f(data.attrs, *args, **kwargs)

        return xr.Dataset(new_vars, data.coords, new_root_attrs)

    return g


def lift_spectrum(function: Callable) -> Callable:
    """
    If the decorated function is supplied a Dataset by the user, this will lift the
    spectrum DataArray and set it as the first argument to the function. The decorated
    functions first argument must be a DataArray.

    Args:
        function: The wrapped function
    """

    @wraps(function)
    def with_dataarray(data: DataType, *args, **kwargs):
        if isinstance(data, xr.Dataset):
            spectrum = data.S.spectrum
            return function(spectrum, *args, **kwargs)
        assert isinstance(data, xr.DataArray), "Data must be a Dataset or DataArray."
        return function(data, *args, **kwargs)

    return with_dataarray


def rename_dataset_keys(dataset: xr.Dataset, rename_keys: dict):
    """
    Renames all keys in a dataset, including data variables, coordinates, dimensions,
    and attributes.
    """
    existing_data_keys = set.union(
        *[set(key_dict.keys()) for key_dict in [dataset.coords, dataset.data_vars]]
    )
    existing_attr_keys = set(dataset.attrs.keys())

    renameable_data = {}
    renameable_attrs = {}
    for key in rename_keys.keys():
        # J: We prefer to rename data over attributes. We may want to swap this or rename both.
        if key in existing_data_keys:
            renameable_data[key] = rename_keys[key]
        elif key in existing_attr_keys:
            renameable_attrs[key] = rename_keys[key]

    dataset = dataset.rename(renameable_data)
    for k, v in renameable_attrs.items():
        dataset.attrs[v] = dataset.attrs.pop(k)

    return dataset



"""Collection of functions to transform data."""

from scipy.ndimage import rotate
from arpes.utilities import lift_spectrum

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from xarray import DataArray


@lift_spectrum
def rotate_spectrum(spectrum: "DataArray", angle: float) -> "DataArray":
    """Rotate a 2D array by a given `angle`.

    Uses scipys `rotate` function and combines the result with coordinates calculated
    from the original array. It doesn't make sense to use this function

    Args
    ----
    spectrum : DataArray
        The 2D array to rotate.
    angle : float
        The angle to rotate the array by in degrees.
    """

    assert spectrum.ndim == 2, "Can only rotate 2D arrays"

    array = spectrum.values
    rotated_array = rotate(array, angle, reshape=False)
